;;; a-stern-CLP.scm

; Da die Stuecke jeweils nicht in alle Alternativen gleichzeitig eingebaut werden, gilt:
; Die Container-liste muss auch die Liste der noch nicht eingebauten Stuecke fuehren!
; Also besser: zustand ist '(<container-liste> <rest-stuecke>)

; Ein Container: (Fassungsvermoegen (Liste der Stuecke))

(load "../breakpoint.scm")
; Das Programm benoetigt die ausgegliederten Funktionen:
(load "prioritaetswarteschlange.scm")
; mit der Funktion (fuege-alle-ein <wegeliste> vor? <priows>)
; Prioritaetswarteschlange (Container-liste-1 ...)
; Das Praedikat vor? muss dafuer definiert werden:
(define
  (vor? zustand-1 zustand-2)
  (< (gesamtkosten zustand-1) (gesamtkosten zustand-2)))

;;; ===== Gesamtkosten =====
; Gesamtkosten sind die Kosten zuzueglich Restkosten.
(define
  (gesamtkosten zustand)
  (+
   (kosten (first zustand))
   (apply + (second zustand)))) ; die Restkosten
;;; ===== Kosten =====
; Kosten sind die Groessen von Containern, die nicht mehr leer sind
(define
  (kosten alle-container)
  (cond
    ((null? alle-container) 0)
    ((null? (second (first alle-container)))
     (kosten (rest alle-container)))
    (else
     (+
      (first (first alle-container))
      (kosten (rest alle-container))))))
   
;;; ===== Restkosten =====
; Restkosten sind hier einfach die restlichen Stuecke
; s.o.: nothing ToDo

; Tests:
;(gesamtkosten '(((60 (30)) (50 ()) (50 ())) (20 20 20 10)))
;(gesamtkosten '(((60 ()) (50 ()) (50 ())) (30 20 20 20 10)))
;(gesamtkosten '(((60 (30 20)) (50 (20)) (50 ())) (20 10)))
;(gesamtkosten '(((60 (30)) (50 (20)) (50 (20))) (20 10)))
;(gesamtkosten '(((60 ()) (50 (30 20)) (50 (20))) (20 10)))

;;; ===== passt? =====
; Hilfsfunktion testet, ob das Stueck noch in den Container passt
(define
  (passt? stueck container)
  (<=
   (+ stueck (apply + (second container)))
   (first container)))

;;; ===== fuelle =====
; Hilfsfunktion zum Einfuellen eines Stuecks in den n-ten Container
(define
  (fuelle container-liste stueck container-nummer) ; 0 basierend
  (define
    (fuelle-ein container stueck)
    (cons (first container) ; Kapazitaet
          (list (cons stueck (second container))) ; stuecke-liste
          ))
  (cond
    ((null? container-liste) (error "leere container-liste" container-liste))
    ((zero? container-nummer)
     (cons
      (fuelle-ein (first container-liste) stueck)
      (rest container-liste)))
    (else
     (cons
      (first container-liste)
      (fuelle (rest container-liste) stueck (- container-nummer 1))))))
; Tests:
;(fuelle '((60 ()) (50 ()) (50 ())) 30 0)
;(fuelle '((60 ()) (50 ()) (50 ())) 30 1)
;(fuelle '((60 ()) (50 ()) (50 ())) 30 2)
     
;;; ===== erstelle-container-liste =====
; Hilfsfunktion, die zu einer Kapazitaetenliste die Liste der leeren
; Container erstellt.
(define
  (erstelle-container-liste kapazitaeten-liste)
  (cond
    ((null? kapazitaeten-liste) ())
    (else
     (cons
      (list (first kapazitaeten-liste) '())
      (erstelle-container-liste (rest kapazitaeten-liste))))))
;Test:
;(erstelle-container-liste '(60 50 50))

;;; ===== Alternativen =====
; bestimmt die Liste der Moeglichkeiten, das uebergebene Stueck in die
; Container der Container-Liste einzufuegen.
(define
  (alternativen zustand)
  (define
    (intern-alternativen container-nummer akku) ; 0 basierend
    (cond
      ((negative? container-nummer)
       akku)
      ((passt?
        (first (second zustand))
        (list-ref (first zustand) container-nummer))
       (intern-alternativen
        (- container-nummer 1)
        (cons
         (list
          (fuelle (first zustand) (first (second zustand)) container-nummer) ; neue Container-Liste
          (rest (second zustand)))                          ; Reststuecke
         akku)))
      (else (intern-alternativen (- container-nummer 1) akku))))
  (intern-alternativen (- (length (first zustand)) 1) '()))

;;; ===== suche =====
(define
  (suche kapazitaeten-Liste alle-stuecke)
  (define
    (suche-intern  priows)
    (ausgabe priows #\newline)
    (cond
      ((null? priows) #f) ; keine Loesung moeglich
      ((null? (second (first priows))) (first (first priows)))   ; Loesung!
      (else
       (suche-intern
        (fuege-alle-ein
         (alternativen (first priows))
         vor?
         (rest priows))))
      ))
  (suche-intern       
   (fuege-alle-ein
    (alternativen
     (list
      (erstelle-container-liste kapazitaeten-Liste)
      alle-stuecke))
    vor?
    '())))


;;; ----- Tests -----

"Tests: erstes Beispiel:"
(define Kapazitaets-Liste-1 '(60 50 50))
(define alle-stuecke '(30 20 20 20 10))
(define leere-Container (erstelle-container-liste Kapazitaets-Liste-1))
Kapazitaets-Liste-1
leere-Container
alle-stuecke
(suche Kapazitaets-Liste-1 alle-stuecke)
;;;"-----------------"
;;"zweites Beispiel:"
;(define Kapazitaets-Liste-2 '(60 60 50))
;(define alle-stuecke '(30 20 20 20 10))
;(define leere-Container (erstelle-container-liste Kapazitaets-Liste-2))
;Kapazitaets-Liste-2
;leere-Container
;alle-stuecke
;(suche Kapazitaets-Liste-2 alle-stuecke)
;;;"-----------------"
;;"drittes Beispiel:"
;(define Kapazitaets-Liste-3 '(60 40 50))
;(define alle-stuecke '(30 20 20 20 10))
;(define leere-Container (erstelle-container-liste Kapazitaets-Liste-3))
;Kapazitaets-Liste-3
;leere-Container
;alle-stuecke
;(suche Kapazitaets-Liste-3 alle-stuecke)
;;"-----------------"
;"viertes Beispiel:"
;(define Kapazitaets-Liste-4 '(90 60 40))
;(define alle-stuecke '(50 40 30 20 20 20 10))
;(define leere-Container (erstelle-container-liste Kapazitaets-Liste-4))
;Kapazitaets-Liste-4
;leere-Container
;alle-stuecke
;(suche Kapazitaets-Liste-4 alle-stuecke)
;;"-----------------"
;;"fuenftes Beispiel:"
;(define Kapazitaets-Liste-5 '(100 100 80 80))
;(define alle-stuecke '(60 40 40 30 30 20 20 20 20 20 20 20 20))
;(define leere-Container (erstelle-container-liste Kapazitaets-Liste-5))
;Kapazitaets-Liste-5
;leere-Container
;alle-stuecke
;(suche Kapazitaets-Liste-5 alle-stuecke)
;;"-----------------"
;"sechstes Beispiel:"
;(define Kapazitaets-Liste-6 '(100 100 80 60))
;(define alle-stuecke '(55 50 40 30 25 20 20 10 10 5 5))
;(define leere-Container (erstelle-container-liste Kapazitaets-Liste-6))
;Kapazitaets-Liste-6
;leere-Container
;alle-stuecke
;(suche Kapazitaets-Liste-6 alle-stuecke)
